/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.geronimo;

import java.io.InterruptedIOException;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Date;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import java.net.InetSocketAddress;
import org.apache.catalina.CometEvent;
import org.apache.catalina.CometProcessor;

import org.apache.http.HttpEntity;
import org.apache.http.HttpException;
import org.apache.http.HttpHost;
import org.apache.http.HttpRequest;
import org.apache.http.HttpResponse;
import org.apache.http.impl.DefaultConnectionReuseStrategy;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.impl.nio.DefaultClientIOEventDispatch;
import org.apache.http.impl.nio.reactor.DefaultConnectingIOReactor;
import org.apache.http.message.BasicHttpRequest;
import org.apache.http.nio.NHttpConnection;
import org.apache.http.nio.protocol.BufferingHttpClientHandler;
import org.apache.http.nio.protocol.EventListener;
import org.apache.http.nio.protocol.HttpRequestExecutionHandler;
import org.apache.http.nio.reactor.ConnectingIOReactor;
import org.apache.http.nio.reactor.IOEventDispatch;
import org.apache.http.nio.reactor.IOReactorException;
import org.apache.http.nio.reactor.SessionRequest;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.apache.http.params.CoreProtocolPNames;
import org.apache.http.protocol.BasicHttpProcessor;
import org.apache.http.protocol.HttpContext;
import org.apache.http.protocol.ExecutionContext;
import org.apache.http.protocol.RequestConnControl;
import org.apache.http.protocol.RequestContent;
import org.apache.http.protocol.RequestExpectContinue;
import org.apache.http.protocol.RequestTargetHost;
import org.apache.http.protocol.RequestUserAgent;
import org.apache.http.util.EntityUtils;

/**
 * Servlet implementation class for Servlet: AsyncServlet
 * 
 * @web.servlet name="NIOServlet" display-name="NIOServlet"
 * 
 * @web.servlet-mapping url-pattern="/nio"
 * 
 */
public class NIOServlet extends HttpServlet implements CometProcessor {
    static final long serialVersionUID = 1L;

    String remoteUrl;
    private static ConnectingIOReactor ioReactor;

    public void init() throws ServletException {
        getServletContext().log("getinit init");
        // Get the value of an initialization parameter
        remoteUrl = getServletConfig().getInitParameter("remoteUrl");

        HttpParams httpParams = new BasicHttpParams(null);
        httpParams
        .setIntParameter(CoreConnectionPNames.SO_TIMEOUT, 5000)
        .setIntParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 10000)
        .setIntParameter(CoreConnectionPNames.SOCKET_BUFFER_SIZE, 8 * 1024)
        .setBooleanParameter(CoreConnectionPNames.STALE_CONNECTION_CHECK, false)
        .setBooleanParameter(CoreConnectionPNames.TCP_NODELAY, true)
        .setParameter(CoreProtocolPNames.USER_AGENT, "HttpComponents/1.1");

        BasicHttpProcessor httpproc = new BasicHttpProcessor();
        httpproc.addInterceptor(new RequestContent());
        httpproc.addInterceptor(new RequestTargetHost());
        httpproc.addInterceptor(new RequestConnControl());
        httpproc.addInterceptor(new RequestUserAgent());
        httpproc.addInterceptor(new RequestExpectContinue());

        BufferingHttpClientHandler handler = new BufferingHttpClientHandler(httpproc, new MyHttpRequestExecutionHandler(),
                                                                                new DefaultConnectionReuseStrategy(), httpParams);

        handler.setEventListener(new EventLogger());

        final IOEventDispatch ioEventDispatch = new DefaultClientIOEventDispatch(handler, httpParams);

        try {
            ioReactor = new DefaultConnectingIOReactor(20, httpParams);
        }
        catch ( IOReactorException e ) {
            throw new ServletException(e);
        }

        Thread t = new Thread(new Runnable() {

                public void run() {
                    try {
                        System.out.println("Running...");
                        ioReactor.execute(ioEventDispatch);
                    }
                    catch ( InterruptedIOException ex ) {
                        System.err.println("Interrupted");
                    }
                    catch ( IOException e ) {
                        System.err.println("I/O error: " + e.getMessage());
                    }
                    System.out.println("Shutdown");
                }
            });

        t.start();
    }

    @Override
    public void destroy() {
        // TODO Auto-generated method stub
        super.destroy();
        System.out.println("Shutting down I/O reactor");
        try {
            ioReactor.shutdown();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * Process the given Comet event.
     * 
     * @param event
     *            The Comet event that will be processed
     * @throws IOException
     * @throws ServletException
     */

    public void event(CometEvent event) throws IOException, ServletException {
        HttpServletRequest request = event.getHttpServletRequest();

        String sessionId = request.getSession(true).getId();
        System.out.println("");
        System.out.println("Session id: " + sessionId + " "
                           + event.getEventType().toString());

        if ( event.getEventType() == CometEvent.EventType.BEGIN ) {
            event.getHttpServletRequest().setAttribute("org.apache.tomcat.comet.timeout", new Integer(100 * 1000));
            event.setTimeout(100 * 1000);

            if ( request.getParameter("remoteUrl") != null ) {
                remoteUrl = request.getParameter("remoteUrl");        
            }

            String remoteUri = request.getParameter("remoteUri");
            if ( remoteUri == null ) {
                remoteUri = "/remoteApp/page";
            }

            Integer[] params = Utils.generateRandomParams();
            System.out.println("sessionId:" + sessionId + "; delay:" + params[0]
                               + "; code:" + params[1]);

            String url_string = remoteUri + "?sessionId=" + sessionId + "&delay=" + params[0] + "&code=" + params[1];       

            
            SessionRequest reqs = ioReactor.connect(new InetSocketAddress(remoteUrl, 8080), null, new MyAttachment(event, sessionId, url_string), null);

            System.out.println("event BEGIN processed");
        }
        else if ( (event.getEventType() == CometEvent.EventType.ERROR)  || (event.getEventType() == CometEvent.EventType.END) ) {
            event.close();
        }

        

    }

    

    static class MyHttpRequestExecutionHandler implements HttpRequestExecutionHandler {

        private final static String TIME               = "time";
        private final static String COMET              = "comet-event";
        private final static String SESSION_ID         = "session-id";
        private final static String REMOTE_URI         = "remote-uri";
        private final static String REQUEST_SENT       = "request-sent";
        private final static String RESPONSE_RECEIVED  = "response-received";


        public void initalizeContext(final HttpContext context, final Object attach) {
            System.out.println("Initializing Context");
            MyAttachment attachment = (MyAttachment)attach;
            context.setAttribute(REMOTE_URI, attachment.getUri());
            context.setAttribute(COMET, attachment.getCometEvent());
            context.setAttribute(SESSION_ID, attachment.getSessionId());
            context.setAttribute(TIME,new Date().getTime());
        }

        public void finalizeContext(final HttpContext context) {
            System.out.println("Finalizing Context");
            Object flag = context.getAttribute(RESPONSE_RECEIVED);
            if ( flag == null ) {
                // Signal completion of the request execution
                /*
                try {
                    ((CometEvent)context.getAttribute(COMET)).close();
                }
                catch (Exception e) {
                    e.printStackTrace();
                }
                */
            }
        }

        public HttpRequest submitRequest(final HttpContext context) {
            System.out.println("Submitting request");
            Object flag = context.getAttribute(REQUEST_SENT);
            if ( flag == null ) {
                // Stick some object into the context
                context.setAttribute(REQUEST_SENT, Boolean.TRUE);

                /*
                System.out.println("--------------");
                System.out.println("Sending request with sessionId: " + (String)context.getAttribute(SESSION_ID));
                System.out.println("--------------");
                */

                return new BasicHttpRequest("GET", (String)context.getAttribute(REMOTE_URI));
            }
            else {
                // No new request to submit
                return null;
            }
        }

        public void handleResponse(final HttpResponse res, final HttpContext context) {
            System.out.println("Handle response");
            HttpEntity entity = res.getEntity();
            try {
                CometEvent event = (CometEvent)context.getAttribute(COMET);
                String content = EntityUtils.toString(entity);
                System.out.println("--------------");
                System.out.println(res.getStatusLine());
                System.out.println("--------------");
                System.out.println("Document length: " + content.length());
                System.out.println("--------------");

                HttpServletResponse response = event.getHttpServletResponse();
                response.setStatus(res.getStatusLine().getStatusCode());
                response.setContentType("text/html");
                if ( content.length() > 0 ) {
                    response.setContentLength(content.length());
                    OutputStream out = null;
                    try {
                        out = response.getOutputStream();
                        out.write(content.getBytes());
                        out.flush();
                        out.close();
                    }
                    catch ( IOException e ) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }

                long initTime = ((Long)context.getAttribute(TIME)).longValue();
                System.out.println("  Responded to " + (String)context.getAttribute(SESSION_ID) + " in " + (new Date().getTime() - initTime));

                event.close();
            }
            catch ( IOException ex ) {
                System.err.println("I/O error: " + ex.getMessage());
            }

            context.setAttribute(RESPONSE_RECEIVED, Boolean.TRUE);

        }

    }

    static class EventLogger implements EventListener {

        public void connectionOpen(final NHttpConnection conn) {
            System.out.println("Connection open: " + conn);
        }

        public void connectionTimeout(final NHttpConnection conn) {
            System.out.println("Connection timed out: " + conn);
        }

        public void connectionClosed(final NHttpConnection conn) {
            System.out.println("Connection closed: " + conn);
        }

        public void fatalIOException(final IOException ex, final NHttpConnection conn) {
            System.err.println("I/O error: " + ex.getMessage());
        }

        public void fatalProtocolException(final HttpException ex, final NHttpConnection conn) {
            System.err.println("HTTP error: " + ex.getMessage());
        }

    }


    static class MyAttachment {
        private String sessionId;
        private String uri;
        private CometEvent event;

        public MyAttachment(CometEvent event, String sessionId, String uri) {
            this.sessionId = sessionId;
            this.uri = uri;
            this.event=event;
        }

        public String getSessionId(){
            return this.sessionId;
        }

        public CometEvent getCometEvent(){
            return this.event;
        }

        public String getUri() {
            return this.uri;
        }
    }
  
    
      
}
